/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.sim.sim.datengenerator;
 
import java.util.Collection;
import java.util.ConcurrentModificationException;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;

import de.bsvrz.dav.daf.main.Data;
import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.OneSubscriptionPerSendData;
import de.bsvrz.dav.daf.main.ResultData;
import de.bsvrz.dav.daf.main.SenderRole;
import de.bsvrz.dav.daf.main.Data.TimeValue;
import de.bsvrz.dav.daf.main.archive.ArchiveData;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.AttributeGroupUsage.Usage;
import de.bsvrz.sys.funclib.debug.Debug;
import de.bsvrz.vew.sim.sim.simulation.OfflineSimulation;
import de.bsvrz.vew.sim.sim.simulation.Simulation;

/**
 * Simulations-Datengenerator für Offline-Simulationen
 *
 * @version $Revision: 1.4 $ / $Date: 2008/12/03 14:00:32 $ / ($Author: Sans $)
 *
 * @author Dambach-Werke GmbH
 * @author Stefan Sans
 *
 */
public class DatenGeneratorOfflineSimulation extends DatenGeneratorSimulation
{
  /**
   * Debug-Logger für Logging-Ausgaben
   */
  private final static Debug _debug = Debug.getLogger();

  /**
   * (Pseudo-) Queue der Eingangsdaten - wird vom {@link #_archivAdapter}
   * gefüllt und vom {@link #_queueArbeiter} geleert
   */
  private LinkedList<ArchiveData> _zuSendendeDaten = new LinkedList<ArchiveData>();

  /**
   * Alle bearbeitenden Datenspezifikationen
   */
  private Map<String, DatenSpezifikation> _sendeAnmeldungen;

  /**
   * Ist der Datengenerator gestartet?
   */
  private boolean _isGestartet = false;

  /**
   * Archiv-Bearbeitungs-Thread
   */
  private Thread _archivAdapter = null;

  /**
   * Eingangsdaten-Bearbeitungs-und Sendethread
   */
  private Thread _queueArbeiter = null;

  /**
   * Beginn der zugeordneten Simulation
   */
  private long _simulationsBeginn;

  /**
   * Ende der zugeordneten Simulation
   */
  private long _simulationsEnde;

  /**
   * Letzter Pausenzeitpunkt - in Echtzeit
   */
  private long _lastPauseReal;

  /**
   * Letzter Pasenzeitpunkt - in Simulationszeit
   */
  private long _lastPauseSimu;

  /**
   * Letzter verabeitetet Zeitstempel
   */
  private long _letzterZeitStempel = 0;

  /**
   * Geschwindigkeitsfaktor der zugeordneten Simulation
   */
  private float _faktor;

  /**
   * Ist der Datengenerator im Zustand "Pause"
   */
  private boolean _isPausiert = false;

  /**
   * Ist der Verarbeitungsthread gerade inaktiv?
   */
  private boolean _isInaktiv;

  /**
   * Plaziere Datensatz in der Pseudo-Queue der Eingangsdaten
   *
   * @param objekt Objekt
   * @param atg Attributgruppe
   * @param asp Aspekt
   * @param zeitStempel Zeitstempel
   * @param data Datensatz
   */
//  void putData(SystemObject objekt, AttributeGroup atg, Aspect asp,
//      long zeitStempel, Data data)
//  {
//    Aspect aspekt = asp;
//    if (atg.isParameter() && asp.getPid().equals("asp.parameterSoll"))
//    {
//      aspekt = getConnection().getDataModel().getAspect("asp.parameterVorgabe");
//    }
//    DataDescription dd = new DataDescription(atg, aspekt,
//        getSimulation().getSimulationsVariante());
//    ResultData rd = new ResultData(objekt, dd, zeitStempel, data);
//
//    synchronized(_zuSendendeDaten)
//    {
//
//       _zuSendendeDaten.add(rd);
//    }
//    if (_queueArbeiter != null && _isInaktiv)
//    {
//      synchronized(_queueArbeiter)
//      {
//        _queueArbeiter.notify();
//      }
//    }
//  }

  /**
   * Plaziere Datensatz in der Pseudo-Queue der Eingangsdaten
   *
   * @param archiveData Archiv-Datensatz der Eingangsdaten
   */
  void putData(ArchiveData archiveData)
  {
//    Aspect aspekt = asp;
//    if (atg.isParameter() && asp.getPid().equals("asp.parameterSoll"))
//    {
//      aspekt = getConnection().getDataModel().getAspect("asp.parameterVorgabe");
//    }
//    DataDescription dd = new DataDescription(atg, aspekt,
//        getSimulation().getSimulationsVariante());
//    ResultData rd = new ResultData(objekt, dd, zeitStempel, data);
//

    if (archiveData.getArchiveTime() >= _simulationsBeginn &&
        archiveData.getArchiveTime() <= _simulationsEnde)
    {
      synchronized(_zuSendendeDaten)
      {
        _zuSendendeDaten.add(archiveData);
      }
      if (_queueArbeiter != null && _isInaktiv)
      {
        synchronized(_queueArbeiter)
        {
          _queueArbeiter.notify();
        }
      }
    }
  }



  /**
   * Bestimme den Beginn der zugeordneten Simulation
   * @return den Simulationsbeginn
   */
  long getSimulationsBeginn()
  {
    return _simulationsBeginn;
  }

  /**
   * Bestimme das Ende der zugeordneten Simulation
   * @return das Simulationsende
   */
  long getSimulationsEnde()
  {
    return _simulationsEnde;
  }

  /**
   * Erzeuge neue Instanz des Datengenerators
   * @param simulation Die Offline-Simulation, für die der Datengenerator
   * erzeugt werden soll
   */
  public DatenGeneratorOfflineSimulation(OfflineSimulation simulation)
  {
    super(simulation);
  }

  @Override
  protected OfflineSimulation getSimulation()
  {
    return (OfflineSimulation)super.getSimulation();
  }

  @Override
  public boolean isGestartet()
  {
    return _isGestartet;
  }

  /**
   * Bestimmt den Pause-Status des Datengenerators
   * @return Ist der Datengenerator im Zustand "Pause"?
   */
  public boolean isPausiert()
  {
    return _isPausiert;
  }

  @Override
  protected void init()
  {
    Collection<DatenSpezifikation> eingangsDaten = analysiereEingangsDaten();
    _archivAdapter = new ArchivAdapter(eingangsDaten, this);
    _queueArbeiter = new QueueArbeiter();

    synchronized(this)
    {
      _simulationsBeginn = getSimulation().getSteuerung().getBeginn();
      _simulationsEnde = getSimulation().getSteuerung().getEnde();

      _lastPauseReal = System.currentTimeMillis();
      _lastPauseSimu = _simulationsBeginn;

      _faktor = getSimulation().getSteuerung().getFaktor();
      _isPausiert = false;
    }
  }

  /**
   * Setze den Datengenerator in den Zustand "Pause"
   */
  public synchronized void pause()
  {
    if (!isPausiert())
    {
        _lastPauseSimu = getSimulationsZeit();
        _isPausiert = true;
    }
    else // Einzelschritt
    {
      ArchiveData ad = null;
      boolean gefunden = false;
      Iterator<ArchiveData> it = _zuSendendeDaten.iterator();

      while (!gefunden && it.hasNext())
      {
        try
        {
          ad = it.next();
          gefunden = ad.getArchiveTime() > _letzterZeitStempel;
        }
        catch (ConcurrentModificationException e)
        {
          it = _zuSendendeDaten.iterator();
        }
      }

      if (gefunden)
      {
        _lastPauseSimu = ad.getArchiveTime();
        if (_queueArbeiter != null && _isInaktiv)
        {
          synchronized(_queueArbeiter)
          {
            _queueArbeiter.notify();
          }
        }
        /****************************************/
        //this.sendeSimulationZeit(_lastPauseSimu);
      }
      else
      {
        _debug.info("Keine Daten für Einzelschritt vorhanden für Simulation " +
            getSimulation());
      }
    }
  }

  @Override
  public synchronized void start()
  {
    if (!isGestartet())
    {
      init();
      _isGestartet = true;
      _archivAdapter.start();
      _queueArbeiter.start();
    }
    else if(isPausiert())
    {
      _isPausiert = false;
      _letzterZeitStempel = this._lastPauseSimu;
      _lastPauseReal = System.currentTimeMillis();

      if (_queueArbeiter != null && _isInaktiv)
      {
        synchronized(_queueArbeiter)
        {
          _queueArbeiter.notify();
        }
      }
    }
    _faktor = getSimulation().getSteuerung().getFaktor();
  }

  @Override
  public void stop()
  {
    if (isGestartet())
    {
      _isGestartet = false;
      _archivAdapter = null;
      _queueArbeiter = null;

      abmeldenSenden();
      _sendeAnmeldungen = null;
      _zuSendendeDaten.clear();
      System.gc();
    }
  }

  /**
   * Sendet eine Zeitangabe als Simulationszeit an den Datenverteiler
   *
   * @param time die Zeit, die als Simualtionszeit gesendet werden soll
   */
  private void sendeSimulationZeit(long time)
  {
    _debug.fine("Sende Simulationszeit " + time);

    AttributeGroup atg = getConnection().getDataModel().getAttributeGroup("atg.simulationsZeit");
    Aspect asp = getConnection().getDataModel().getAspect("asp.standard");

    String key = DatenSpezifikation.erzeugeSchluessel(atg, asp);

    DatenSpezifikation ds = _sendeAnmeldungen == null ? null : _sendeAnmeldungen.get(key);
    if (ds == null)
    {
      boolean abmeldenErlaubt = true;
      ds = new DatenSpezifikation(atg, asp, (short)0, getSimulation().getObjekt());

      try
      {
        getConnection().subscribeSender(this, getSimulation().getObjekt(),
            ds, SenderRole.source());
      }
      catch (OneSubscriptionPerSendData e)
      {
        abmeldenErlaubt = false;
      }

      if (abmeldenErlaubt)
      {
        if (_sendeAnmeldungen == null)
        {
          _sendeAnmeldungen = new HashMap<String, DatenSpezifikation>();
        }
        _sendeAnmeldungen.put(key, ds);
      }
    }

    try
    {
      Data data = getConnection().createData(atg);
      TimeValue timeValue = data.getTimeValue("Zeit");
      timeValue.setMillis(time);

      ResultData rd = new ResultData(getSimulation().getObjekt(), ds,
          System.currentTimeMillis(), data);
      super.sendeDaten( new ResultData[] { rd });
    }
    catch (RuntimeException e)
    {
      _debug.error("Fehler beim Senden der Simulationszeit");
    }
  }

  /**
   * Bestimmt die Simulationszeit
   *
   * @return die Simulationszeit in ms seit 1.1.1970, gemäss dem
   * Simulationszeitfluss und unter Berücksichtigung der Simulations-
   * Geschwindigkeit und des Zustands "Pause"
   */
  private long getSimulationsZeit()
  {
    if (isPausiert())
    {
      return _lastPauseSimu;
    }
    else
    {
      return Math.round((double)_lastPauseSimu + (_faktor *
          (double)(System.currentTimeMillis() - _lastPauseReal)));

    }
  }

  /**
   * Abmelden aller Sendeanmeldungen
   */
  private void abmeldenSenden()
  {
    if (_sendeAnmeldungen != null)
    {
      _debug.fine("Abmelden von " + _sendeAnmeldungen.size() +
          " SendeAnmeldungen für Simulation " + getSimulation());
      for (DatenSpezifikation ds: _sendeAnmeldungen.values())
      {
        try
        {
          getConnection().unsubscribeSender(this, ds.getSystemObjekte(), ds);
        }
        catch (Exception e)
        {
          _debug.error("Fehler beim abmelden von "  + ds + " für Simulation " +
              getSimulation(), e);
        }
      }
      _sendeAnmeldungen.clear();
    }
  }

  @Override
  protected void sendeDaten(Collection<ResultData> sendePuffer)
  {
    Map<String, DatenSpezifikation> lokaleAnmeldung =
      new HashMap<String, DatenSpezifikation>();

    for (ResultData rd: sendePuffer)
    {
      AttributeGroup atg = rd.getDataDescription().getAttributeGroup();
      Aspect asp = rd.getDataDescription().getAspect();
      String key = DatenSpezifikation.erzeugeSchluessel(atg, asp);
      DatenSpezifikation ds = _sendeAnmeldungen.get(key);
      if (ds == null || !ds.getSystemObjekte().contains(rd.getObject()))
      {
        DatenSpezifikation dsLokal = lokaleAnmeldung.get(key);
        if (dsLokal != null)
        {
          dsLokal.getSystemObjekte().add(rd.getObject());
        }
        else
        {
          dsLokal = new DatenSpezifikation(atg, asp,
              getSimulation().getSimulationsVariante(), rd.getObject());
          lokaleAnmeldung.put(key, dsLokal);
        }
      }
    }

    for (DatenSpezifikation dsLokal: lokaleAnmeldung.values())
    {
      boolean abmeldenErlaubt = true;

//      SenderRole role = SenderRole.sender();
      SenderRole role = SenderRole.source();;
      AttributeGroup atg = dsLokal.getAttributeGroup();
      Aspect asp = dsLokal.getAspect();
      AttributeGroupUsage atgv = atg.getAttributeGroupUsage(asp);
      Usage benutzung = atgv.getUsage();
//      if (benutzung == Usage.OnlineDataAsSourceReceiver)
//      {
//        role = SenderRole.source();
//      }
      if (benutzung == Usage.OnlineDataAsSenderDrain)
      {
        role = SenderRole.sender();
      }

      try
      {
        getConnection().subscribeSender(this, dsLokal.getSystemObjekte(), dsLokal, role);
      }
      catch (Exception e)
      {
        // Abmeldung darf nicht von hier erfolgen
        abmeldenErlaubt = false;
        if (!(e instanceof OneSubscriptionPerSendData))
        {
          _debug.error("Sendeanmeldung mit " + dsLokal + " für Simulation " +
              getSimulation() + " fehlegeschlagen", e);
        }
      }
      if (abmeldenErlaubt)
      {
        String key = DatenSpezifikation.erzeugeSchluessel(atg, asp);
        DatenSpezifikation ds = _sendeAnmeldungen.get(key);
        if (ds == null)
        {
          _sendeAnmeldungen.put(key, dsLokal);
        }
        else
        {
          ds.getSystemObjekte().addAll(dsLokal.getSystemObjekte());
        }
      }
    }

    super.sendeDaten(sendePuffer);
  }

  /**
   * Asynchrones Abarbeiten und Versenden der Daten, die aus dem Archiv hereinkommen
   *
   * @author Stefan Sans
   */
  public class QueueArbeiter extends Thread
  {
    /**
     * Maximale Anzahl Datensätze in einem Sendepuffer
     */
    final static int MAX_BUFFER = 1000;

    /**
     * Maximale Zeit der Inaktivität in ms
     */
    final static int MAX_SLEEP = 500;

    /**
     * Mindestabstand zu Datensätzen aus dem nächsten Durchlauf, damit
     * Inaktivität eintreten kann
     */
    final static int TIME_SLOT = 10;

    /**
     * Einstiegspunkt für die Archivdaten-Bearbeitung
     * 
     * * @see java.lang.Runnable#run()
     */    
    public void run()
    {
      _letzterZeitStempel = 0;
      
      while(isGestartet())
      {
        long aktSimuZeit = getSimulationsZeit();
        
        // Sende SimulationsZeit initial 1x
        if (_letzterZeitStempel == 0)
        {
          sendeSimulationZeit(aktSimuZeit);
          _letzterZeitStempel = aktSimuZeit;
        }
        
        // Gehe in den Zustand Stop, wenn das zeitliche Ende erreicht ist
        // UND
        // alle Daten verarbeitet sind
        if (aktSimuZeit >= _simulationsEnde)
        {
          if (_zuSendendeDaten.isEmpty() && (_archivAdapter == null ||
              _archivAdapter.getState() == Thread.State.TERMINATED))
          {  
            _debug.info("Simulation " + getSimulation() + " wird terminiert, da ihr Ende erreicht ist");
            getSimulation().setZustand(Simulation.ZustandsBezeichner.Stop.text);
            break;
          }  
        }
        
        List<ResultData> sendePuffer = null;
        long pufferZeitStempel = 0;
        long naechsterZeitStempel = 0;
        ListIterator<ArchiveData> it = _zuSendendeDaten.listIterator();
        
        while((sendePuffer == null || sendePuffer.size() < MAX_BUFFER) &&
            it.hasNext())
        {
          try
          {
            ArchiveData ad = it.next();
            aktSimuZeit = getSimulationsZeit();
            long zeitStempel = ad.getArchiveTime();
           
            if (zeitStempel <= aktSimuZeit)
            {
              if (zeitStempel > pufferZeitStempel)
              {
                pufferZeitStempel = zeitStempel;
              }
              if (sendePuffer == null)
              {
                sendePuffer = new LinkedList<ResultData>();        
              }
              
              AttributeGroup atg = ad.getDataDescription().getAttributeGroup();
              Aspect asp = ad.getDataDescription().getAspect();
              if (atg.isParameter() && asp.getPid().equals("asp.parameterSoll"))
              {
                asp = getConnection().getDataModel().getAspect("asp.parameterVorgabe");
              }
              DataDescription dd = new DataDescription(atg, asp, getSimulation().getSimulationsVariante());
              ResultData rd = new ResultData(ad.getObject(), dd, ad.getDataTime(), ad.getData());
              
              sendePuffer.add(rd);
              synchronized(_zuSendendeDaten)
              {
                it.remove();
              }  
            }
            else
            {
              if (naechsterZeitStempel == 0 || 
                  zeitStempel < naechsterZeitStempel)
              {
                naechsterZeitStempel = zeitStempel;
              }
            }
          }
          catch (ConcurrentModificationException e)
          {
            it = _zuSendendeDaten.listIterator();
          }
        }
         
        if (sendePuffer != null)
        {  
          if (pufferZeitStempel > _letzterZeitStempel)
          {
//            sendeSimulationZeit(aktSimuZeit);
            sendeSimulationZeit(pufferZeitStempel);
            _letzterZeitStempel = pufferZeitStempel;
          }
         
          sendeDaten(sendePuffer);
        }
        
        if (_zuSendendeDaten.isEmpty())
        {
          // Forciert maximale inaktive Zeit bei leerer Sendequeue
          naechsterZeitStempel = aktSimuZeit + MAX_SLEEP + TIME_SLOT; 
        }
                
        if (naechsterZeitStempel > aktSimuZeit + TIME_SLOT)
        {
          try
          {
            long inaktivZeit = Math.round((double)(naechsterZeitStempel - 
                aktSimuZeit - TIME_SLOT) / _faktor);
            // Thread.sleep(inaktivZeit);
            if (inaktivZeit > 0)
            {  
              synchronized(this)
              {
                _isInaktiv = true;
                wait(inaktivZeit);
                _isInaktiv = false;
              }
            }  
          }
          catch (InterruptedException e)
          {
          }
        }
      }
      _debug.config("Asynchrone Bearbeitung von Archivdaten terminiert");
    }
  }
  
}
