/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 * 
 * Contact Information:
 * Dambach-Werke GmbH
 * Elektronische Leitsysteme
 * Fritz-Minhardt-Str. 1
 * 76456 Kuppenheim
 * Phone: +49-7222-402-0
 * Fax: +49-7222-402-200
 * mailto: info@els.dambach.de
 */

package de.bsvrz.vew.sim.sim.datengenerator;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import de.bsvrz.dav.daf.main.DataDescription;
import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.SystemObject;

/**
 * Jede Instanz dieser Klasse repräsentiert genau eine Datenspezifikation so,
 * wie sie im Bereich Simulationen bei der Generierung von Simulationsdaten
 * benötigt wird. Im Wesentlichen wird die Klasse <code>DataDescription</code>
 * um eine Menge von Systemobjekten erweitert.
 *
 * @author Dambach Werke GmbH
 * @author Stefan Sans
 * @version $Revision: 1.2 $ / $Date: 2008/10/31 13:30:31 $ / ($Author: Sans $)
 */
public class DatenSpezifikation extends DataDescription
{
  /**
   * Menge von Systemobjekten
   */
  private Set<SystemObject> _systemObjekte = new HashSet<SystemObject>();

  /**
   * Zugriff auf die Menge von Systemobjekten dieser <code>DatenSpezifikation</code>-Instanz
   * @return Die Menge von Systemobjekten
   */
  public Set<SystemObject> getSystemObjekte()
  {
    return _systemObjekte;
  }

  /**
   * Erzeugt eine neue Instanz vom Typ <code>DatenSpezifikation</code>. Die Menge
   * der Systemobjekte ist leer
   * @param atg die Attributgruppe
   * @param asp der Aspekt
   * @param sv die Simulationsvariante
   */
  public DatenSpezifikation(AttributeGroup atg, Aspect asp, short sv)
  {
    super(atg, asp, sv);
  }

  /**
   * Erzeugt eine neue Instanz vom Typ <code>DatenSpezifikation</code> mit einem
   * initialen Objekt in der Menge der Systemobjekte
   * @param atg die Attributgruppe
   * @param asp der Aspekt
   * @param sv die Simulationsvariante
   * @param obj Das initiale Objekt in der Menge der Systemobjekte
   */
  public DatenSpezifikation(AttributeGroup atg, Aspect asp, short sv, SystemObject obj)
  {
    this(atg, asp, sv);

    _systemObjekte.add(obj);
  }

  /**
   * Erzeugt eine neue Instanz vom Typ <code>DatenSpezifikation</code> mit einer
   * initialen Menge von Systemobjekten
   * @param atg die Attributgruppe
   * @param asp der Aspekt
   * @param sv die Simulationsvariante
   * @param col Sammlung von Objekten, die die initiale Menge von Systemobjekten bilden
   */
  public DatenSpezifikation(AttributeGroup atg, Aspect asp, short sv, Collection<SystemObject> col)
  {
    this(atg, asp, sv);

    _systemObjekte.addAll(col);
  }

  /**
   * Erzeugt einen (Text-) Schlüssel, der für die übergebene Attributgruppen-/Aspekt-Kombination eindeutig ist
   * @param atg die Attributgruppe
   * @param asp der Aspekt
   * @return den eindeutigen Textschlüssel
   */
  public static String erzeugeSchluessel(AttributeGroup atg, Aspect asp)
  {
    return atg.getPid() + "\\" + asp.getPid();
  }
  
}
